using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;

namespace RayTracer
{
    class Ellipse : BaseShape
    {
        protected double a;
	protected double b;
	protected double c;
	
	protected Vector3D pos;
	protected int materialIndex;
	
	const int NUM_DATA = 7;
	
	//_vZ
	//E:_ V:xNg
	//ԂlFՓˈʒuiՓ˂ĂȂꍇ-1j
	public double calcT(Vector3D E, Vector3D V)
	{
		double alpha, beta, gamma;	//WA萔
		double D;		//ʎ
		double K = 1;
		Matrix M;
		Vector3D R0d = Vector3D.sub(E, this.pos);	//R0'

		M = Matrix.scaleMatrix(1 / (this.a * this.a),
							   1 / (this.b * this.b),
							   1 / (this.c * this.c));
							   
		alpha = Vector3D.dotProduct(Matrix.tVectorMatrix(V, M), V);
		beta  = Vector3D.dotProduct(Matrix.tVectorMatrix(V, M), R0d);
		gamma = Vector3D.dotProduct(Matrix.tVectorMatrix(R0d, M), R0d) - K;
		
		D = beta * beta  - alpha * gamma;

		if(D < 0)
		{
			return -1;
		}
		else
		{
			return (-beta - Math.Sqrt(D)) / alpha;
		}

	}
	
	//@vZ
	//PF_
	//ԂlF@
	public Vector3D calcNorm(Vector3D P)
	{
		Vector3D n;
		Matrix smat = Matrix.scaleMatrix(1 / (this.a * this.a), 
								  		 1 / (this.b * this.b), 
								  		 1 / (this.c * this.c));
								  
		n = Vector3D.normalize(Matrix.matrixVector(smat, Vector3D.sub(P, this.pos)));

		return n;
	}
	
	//VF[fBOvZ
	//QF_@PF_@EF_
	//mat: gp}eA
	public Color calcShading(PointLight Q ,Vector3D P, Vector3D E, 
	 						Material material)
	{
		double ia, id, s, ii;
		double dot_L_N;
		int pR, pG, pB;
		
		Vector3D N = this.calcNorm(P);
		Vector3D L = Vector3D.sub(Q.getPosition(), P);
		double r = Vector3D.length(L);
		L = Vector3D.normalize(L);
		
		dot_L_N = Vector3D.dotProduct(L, N);
		
		ia = material.getKa();
		
		if (dot_L_N < 0)
		{
			id = s = 0;
		}
		else
		{	
			//gUˌxvZ
			id = (material.getKd() * Q.getIntensity() * dot_L_N) / (r * r);
			
			//ʔˌxvZ
			
			//PʎxNg
			Vector3D V = Vector3D.normalize(Vector3D.sub(E, P));
			
			//PʐˌxNg
			Vector3D R = Vector3D.normalize(Vector3D.sub(Vector3D.scale(2 * Vector3D.dotProduct(N, L), N), L));
			
			double dot_R_V = Vector3D.dotProduct(V, R);
			
			//ʔˌxvZ
			s = (material.getKs() * Q.getIntensity() * Math.Pow(dot_R_V, material.getN())) /
					(r * r);
		}
		
		//ʔˌF
		Color specColor = Color.FromArgb(255, 255, 255);
		
		//̐F
		Color mColor = material.getColor();
		
		pR = mColor.R;
		pG = mColor.G;
		pB = mColor.B;
		
		pR = (int)(pR * ia + pR * id + specColor.R * s + 0.5);
		pG = (int)(pG * ia + pG * id + specColor.G * s + 0.5);
        pB = (int)(pB * ia + pB * id + specColor.B * s + 0.5);
		
		if (pR < 0) pR = 0; else if (pR > 255) pR = 255;
		if (pG < 0) pG = 0; else if (pG > 255) pG = 255;
		if (pB < 0) pB = 0; else if (pB > 255) pB = 255;

        return Color.FromArgb(pR, pG, pB);
	}
	
	//gp}eAԍ擾
	//ԂlF̂̕gp}eAԍ
	public int getUsingMaterialIndex()
	{
		return materialIndex;
	}
	
	//RXgN^
	public Ellipse()
	{
		a = 0;
		b = 0;
		c = 0;
		
		pos = new Vector3D();
	}
	
	//lwRXgN^
	public Ellipse(double a, double b, double c, 
				   double x, double y, double z)
	{
		this.a = a;
		this.b = b;
		this.c = c;
		
		this.pos = new Vector3D(x, y, z);
	}
	
	//f[^ꂽŏ
	//ȉ~f[^tH[}bg:x y z a b c material_index
	public Ellipse(String data)
	{
		double x, y, z;
	
		try
		{
            string[] delimiter = {" "};
            string[] s = data.Split(delimiter, StringSplitOptions.None);		//Xy[XƂɕ񕪊
			if (s.Length >= NUM_DATA){						//͂ꂽKvȏȂǂݍ
                x = Double.Parse(s[0]);
                y = Double.Parse(s[1]);
                z = Double.Parse(s[2]);

                a = Double.Parse(s[3]);
                b = Double.Parse(s[4]);
                c = Double.Parse(s[5]);

                materialIndex = Int32.Parse(s[6]);
				
				pos = new Vector3D(x, y, z);
			}
			else
                throw new ArgumentException("[Ellipse]͒ľ܂:" + data);
		}
        catch (ArgumentException e)
		{
			throw e;
		}
	}
    }
}
